#!/usr/local/bin/perl
#
# Copyright (c) 2003-2007 Jeremy Kister
# Author: Jeremy Kister <argus-devel @t jeremykisterc.om>
# Date: 2007-Nov-28 19:15 (EDT)
# Function: send messages via AIM/OSCAR

use strict;
use Net::OSCAR;
use Getopt::Std;

$| = 1; #flush
my $idle_time=0;
my $commit_ok = 1; # 0 pending(not ok), 1=never commited, 2=flagged 0
my $allow_bfs=0;
my (%opt,%buddies,%ids,%hush,$sentaway);
chdir('/') || die "could not chdir /: $!\n";

getopts('u:g:d:s:p:Db:', \%opt);

$SIG{USR1} = sub {
	warn "received SIGUSR1: running queue\n" if($opt{D});
	$idle_time = 0;
};

my $aaim_user = $ENV{'AAIM_USER'} || $opt{s};
my $aaim_pass = $ENV{'AAIM_PASS'} || $opt{p};

unless($aaim_user && $aaim_pass && $opt{b}){
	print "syntax:\n",
	      "	aim.pl [-u <username>] [-g <group>] [-D] [-d </path/argus/sbin>] -s <screenname> -p <password> -b <Buddies>\n",
	      "	options:\n",
	      "		-u	username to run as\n",
 	      "		-g	group to run as\n",
	      "		-D	print debugging info\n",
	      "		-d	alternative path to argus sbin directory (default: /usr/local/sbin)\n",
	      "		-s	AIM screenname (or set AAIM_USER in environment)\n",
	      "		-p	AIM password (or set AAIM_PASS in environment)\n",
	      "		-b	list of buddies (separate multiples by comma, no spaces in usernames)\n\n";
	exit;
}

# Change to nobody if you'd like, or to a user who has argusctl permissions.
if(defined($opt{g})){
	my $gid = (getgrnam($opt{g}))[2];
	$gid = $opt{g} if( !defined($gid) && $opt{g} =~ /^\d+$/ );
	slowdie("invalid group for -g option. aborting.") unless defined $gid;
	$( = $gid; $) = $gid;
}
if(defined($opt{u})){
	my $uid = (getpwnam($opt{u}))[2];
	$uid = $opt{u} if( !defined($uid) && $opt{u} =~ /^\d+$/ );
	slowdie("invalid user for -u option. aborting.") unless defined $uid;

	# fix /var/aaimqueue/ if owned by anyone other than $opt{u}
	if(-d '/var/aaimqueue'){
		my $diruid = (stat('/var/aaimqueue/'))[4];
		if($uid != $diruid){
			warn "/var/aaimqueue not owned by $opt{u}: fixing..\n";
			chown($uid,$(,'/var/aaimqueue/') || slowdie("cannot chown /var/aaimqueue/: $!");
		}
		# fix /var/aaimqueue/ files if owned by anyone other than $opt{u}, (so we dont keep dying)
		if(opendir(DIR, '/var/aaimqueue/')){
			foreach my $subdir (grep {!/^(?:\.|PID)/} readdir DIR){
				my $sdiruid = (stat("/var/aaimqueue/${subdir}"))[4];
				if($sdiruid != $uid){
					warn "/var/aaimqueue/${subdir} not owned by $opt{u}: fixing..\n";
					chown($uid,$(,"/var/aaimqueue/${subdir}/") || slowdie("cannot chown $subdir/: $!");
				}
					
				if(opendir(SUBDIR, "/var/aaimqueue/$subdir")){
					foreach my $file (grep {!/^\./} readdir SUBDIR){
						if($subdir eq 'tmp'){
							# files are worthless, could be half-written, unknown recips, etc.
							unlink("/var/aaimqueue/tmp/$file") || warn "could not unlink tmp/$file: $!\n";
						}else{
							my $fileuid = (stat("/var/aaimqueue/${subdir}/${file}"))[4];
							if($uid != $fileuid){
								warn "/var/aaimqueue/${subdir}/${file} not owned by $opt{u}: fixing..\n";
								chown($uid,$(,"/var/aaimqueue/${subdir}/${file}") || slowdie("cannot chown $file: $!");
							}
						}
					}
					closedir SUBDIR;
				}
			}
			closedir DIR;
		}	
	}else{
		warn "/var/aaimqueue/ does not exist - creating..\n";
		mkdir('/var/aaimqueue',0700) || slowdie("cannot create /var/aaimqueue/: $!");
		chown($uid,$(,'/var/aaimqueue/') || slowdie("cannot chown /var/aaimqueue/: $!");
	}

	# write the PID before we chuid
	if(open(PID, '>/var/aaimqueue/PID')){
		print PID $$;
		close PID;
	}else{
		slowdie("cannot write to /var/aaimqueue/PID: $!");
	}
	$! = 0;
	$< = $> = $uid;
	die "unable to change uid: $!\n" if $!;
}

my $sbin = $opt{d} || '/usr/local/sbin';
$opt{b} =~ s/\s+//g;
my @buddies = split(/,/, lc($opt{b}));

print "Starting Server...\n";

my $aim = Net::OSCAR->new();
# Set up the handlers for commands issued by the server.
$aim->set_callback_signon_done(\&signon_done);
$aim->set_callback_im_in(\&im_in);
$aim->set_callback_im_ok(\&im_ok);
$aim->set_callback_error(\&error);
$aim->set_callback_evil(\&evil);
$aim->set_callback_buddy_info(\&buddy_info);
$aim->set_callback_buddylist_ok(\&buddylist_ok);
$aim->set_callback_buddylist_error(\&buddylist_error);
$aim->set_callback_buddy_in(\&buddy_in);
$aim->set_callback_buddy_out(\&buddy_out);

while(1){
	unless($aim->is_on){
		signon($aim);
	}
	# check to see if we have incoming messages waiting
	$aim->do_one_loop();

	run_queue();

	if($idle_time <= 5){
		if($sentaway){
			print "WHO!? WHA!? -- damn it.. i hate being woken up!\n" if($opt{D});
			$aim->set_away();
			$sentaway=0;
		}
		$idle_time++;
	}elsif($idle_time <= 20){
		sleep 1;
		$idle_time++;
	}else{
		sleep 5;
		$idle_time += 5;
		unless($sentaway){
			if($idle_time > 180){
				print "<yawn> -- taking a nap..\n" if($opt{D});
				$aim->set_away("Taking a nap..");
				$sentaway=1;
			}
		}
	}
}

sub run_queue {
	# check to see if we have outgoing messages waiting
	if(opendir(DIR, '/var/aaimqueue/')){
		foreach my $subdir (grep {!/^(?:\.|PID|tmp)/} readdir DIR){ # $subdir is recipient name
			if(opendir(SUBDIR, "/var/aaimqueue/$subdir/")){
				my $sendmsg;
				foreach my $file (grep {!/^(?:\.|tmp)/} readdir SUBDIR){
					if($buddies{$subdir}){
						if( $hush{$subdir} && ($hush{$subdir} > time()) ){
							print "skipping $subdir (hushed)\n" if($opt{D});
						}else{
							if( $hush{$subdir} && ($hush{$subdir} < time()) ){
								print "hush for $subdir has expired.\n" if($opt{D});
								delete $hush{$subdir};
							}
							if(open(FILE, "/var/aaimqueue/${subdir}/${file}")){
								while(<FILE>){
									$sendmsg .= $_;
								}
								close FILE;
							}
						}
					}else{
						print "skipping $subdir (not online)\n" if($opt{D});
					}
					unless(unlink("/var/aaimqueue/${subdir}/${file}")){
						slowdie("cannot unlink ${subdir}/${file}: $!"); # we dont want to keep sending msg-die to clean
					}
				}
				if($sendmsg && $buddies{$subdir}){
					chomp($sendmsg);
					# stop client from changing things like :Ping to <happy face>ing
					$sendmsg =~ s/:/ -> /g;

					print "$aaim_user -> ${subdir}: ${sendmsg}\n" if($opt{D});
					my $reqid = $aim->send_im($subdir, $sendmsg);
					if($reqid eq '0'){
						warn "message too long: dropped\n";
					}else{
						$ids{$reqid} = 1; # something in Net::OSCAR doesnt give same id as im_ok
					}
					$idle_time=0;
				}
			}else{
				warn "cannot open /var/aaimqueue/${subdir}\n";
			}
		}
		closedir DIR;
	}else{
		warn "could not open /var/aaimqueue/: $!\n";
	}
}

sub signon {
	my $aim = shift;
	warn "about to connect to AIM\n";

	my $i = my $r = 1;
	until($aim->signon(screenname => $aaim_user,
			       		 password	=> $aaim_pass)){
		if($i == 10){
			sleep 60;
			$i=0;
		}else{
			sleep 1;
		}
		warn "attempting signon to aim [$i]\n";
		$i++;
	}
	  

	$i = $r = 1 ;
	until($aim->is_on){
		warn "waiting for signon confirmation [$i/60 - round $r]\n" if($opt{D});
		$aim->do_one_loop();
		if($i == 60){
			warn "tried connecting 60 times.. sleeping 300 seconds\n" if($opt{D});
			sleep 300;
			$i=0;
			$r++;
		}else{
			sleep 1;
		}
		$i++;
	}

	warn "about to add buddies\n" if($opt{D});
	$aim->add_buddy('Buddies',@buddies);
	buddylist_safe_commit();
	$allow_bfs=1;

	warn "about to add_permit\n" if($opt{D});
	$aim->add_permit(@buddies);
	buddylist_safe_commit();
	$allow_bfs=1;

}

sub buddylist_safe_commit {
	warn "see commit_ok: $commit_ok\n" if($opt{D});
	for(my $i=0; $i<60; $i++){
		if($commit_ok){
			warn "have commit_ok, committing buddylist [$i/60].\n" if($opt{D});
			$commit_ok = 2;
			$aim->commit_buddylist();
			$commit_ok = 0 if($commit_ok == 2);  # not if buddy_ok or buddy_error
			return;
		}else{
			warn "waiting for commit_ok [$i/60].\n" if($opt{D});
			sleep 1 if($i % 2);
		}
		$aim->do_one_loop();
	}
	unless($commit_ok){
		warn "buddy list did not commit: sleeping 5.";
		sleep 5;
	}
}

sub signon_done {
	my $self = shift;

	warn "connected to aim server as $aaim_user.\n" if($opt{D});
}

sub buddylist_ok {
	my $self = shift;

	warn "Buddy list OK\n" if($opt{D});
	$commit_ok=1;
}

sub buddylist_error {
	my ($self,$error,$what) = @_;
	
	warn "Buddylist error: $error - what: $what\n";
	$commit_ok=1;
}

sub buddy_info {
	my ($self,$screenname,$data) = @_;

	next unless($allow_bfs);

	if($data->{away}){
		warn "$screenname is away\n" if($opt{D});
	}
	$screenname = lc($screenname);
	$screenname =~ s/\s+//g;
	$buddies{$screenname} = $data->{onsince} || 1;
	warn "buddies{$screenname} = $buddies{$screenname}\n" if($opt{D});
}

sub buddy_in {
	my ($self,$buddy,$group,$data) = @_;

	next unless($allow_bfs);

	$buddy = lc($buddy);
	$buddy =~ s/\s+//g;
	if($data->{online}){
		if(grep /^${buddy}$/, @buddies){
			$buddies{$buddy} = $data->{onsince} || 1;
			warn "IN: $buddy [$group] is online\n" if($opt{D});
		}else{
			# remove the buddy
			$aim->remove_buddy($group,$buddy);
			buddylist_safe_commit();
		   $allow_bfs=1;
		}
	}
}

sub buddy_out {
	my ($self,$buddy,$group) = @_;

	next unless($allow_bfs);

	$buddy = lc($buddy);
	$buddy =~ s/\s+//g;

	warn "$buddy [$group] signed off\n" if($opt{D});
	$buddies{$buddy} = 0;

	# buddy could have been signed on in multiple places
	$aim->get_info($buddy);
}

sub error {
	# called when an error occurs while communicating 
	my ($aim, $conn, $error, $description, $fatal) = @_;

	warn "ERROR: $error - description: $description - fatal: $fatal\n";
	if($error == 0){
		if($description =~ /connecting too frequently/){
			warn "sleeping 1200\n";
			sleep 1200;
		}else{
			# not finished signing on
			warn "sleeping 10\n";
			sleep 10;
		}
	}elsif($error == 4){
		# user not logged on
		next;
	}else{
		warn "sleeping 2\n";
		sleep 2;
	}
	die "detected fatal error\n" if($fatal);
}

sub rate_alert {
	my ($aim,$level,$clear,$window,$worrisome) = @_;

	next if($level eq 'RATE_CLEAR');
	my $sleep = sprintf("%d",($clear * 1000));
	
	warn "ALERT: got level $level -> sleeping $sleep seconds\n";
	sleep $sleep;
}

sub evil {
	# when the bot recieves a warning.
	my ($aim,$newevil,$from) = @_;

	print "EVIL: our warning is now ${newevil}% via [$from].\n";

	#Warn and block him, if its not anonymous
	if($from){
		#Remove spaces, make it lowercase.
		$from = lc($from);
		$from =~ s/\s+//g;

		# force it to be impossible to warn ourselves
		if($from eq $aaim_user){
			warn "wont evil self!\n";
		}else{
			$aim->evil($from, 0);
			$aim->add_deny($from);
			buddylist_safe_commit();
			$allow_bfs=1;
		}
	}
}

sub im_ok {
	my ($aim, $to, $reqid) = @_;
	
	if($ids{$reqid}){
		warn "im_ok: got success on id [$reqid] to [$to]\n" if($opt{D});
		delete $ids{$reqid};
	}else{
		warn "im_ok: got success to unknown id: $reqid to [$to]\n" if($opt{D});
	}
}

sub im_in {
	# called when the bot recieves an IM.
	my ($aim, $sender, $msg, $is_away) = @_;

	$sender =~ s/\s+//g;
	$sender = lc($sender);

	# never get tricked into accepting a message from "ourself"
	if($sender eq $aaim_user){
		warn "INFO: got message from self: [$msg]!\n";
		return;
	}

	#Format the message without HTML.
	$msg =~ s/<(.|\n)+?>//g;

	if($opt{D}){
		print "[AWAY] " if($is_away);
		print "${sender}: ${msg}\n";
	}

	# we can do fun stuff here (is secure as AIM buddy list)
	if($buddies{$sender}){
		my $send;
		if($msg =~ /^ACK:\s*(\d+|all)$/){
			my $pageid=$1;
			$send="Acking PageID: $pageid\n";
			$send .= `$sbin/argusctl notify_ack idno=$pageid AIM::$sender`;
		}elsif($msg =~ /^\/msg\s+(\S+)\s+(.+)/){
			my $resendee = $1;
			my $msgx = "[via $sender]: " . $2;
			my $reqid = $aim->send_im($resendee,$msgx);
			if($reqid eq '0'){
				warn "message too long: dropped\n";
				$send="Message to $resendee too long";
			}else{
				$ids{$reqid} = 1;
				$send="Message resent to $resendee";
			}
		}elsif($msg eq 'argusctl status'){
			$send=`$sbin/argusctl status`;
		}elsif($msg eq 'ping'){
			$send='pong';
		}elsif($msg eq 'whoson'){
			while(my($key,$value) = each %buddies){
				$send .= "$key -> $value\n" if($value);
			}
		}elsif($msg eq 'sleep'){
			warn "sleep command received - taking a nap\n";
			$aim->set_away("Sleeping..");
			$idle_time += 5;
			$sentaway=1;
		}elsif($msg =~ /^HUSH:\s*(-?\d{1,2})$/){
			my $hush_min = $1;
			if($hush_min > 0){
				warn "hush command received: will not notify $sender for $hush_min minutes\n";
				$hush{$sender} = (time() + (${hush_min}*60));
				$send="HUSHed for $hush_min minutes";
			}else{
				warn "hush command received: hush manually expired.\n";
				delete $hush{$sender};
				$send="HUSH manually expired.";
			}
		}
		if(defined($send)){
			my $reqid = $aim->send_im($sender,$send);
			if($reqid eq '0'){
				warn "message too long: dropped\n";
			}else{
				$ids{$reqid} = 1; # bug in Net::OSCAR doesnt give same ID as in im_ok
			}
			print "[$reqid] $aaim_user -> ${sender}: ${send}\n" if($opt{D});
			$idle_time=0; #only reset idle if a valid command was sent to us
		}
	}
}

sub slowdie {
	my $err = shift;
	warn "$err\n";
	sleep 5;
	exit 1;
}

